/* -*- c++ -*- */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <list>
#include <iostream>
#include <string>
#include <vector>


#define USE_WCHAR 1

#ifdef USE_WCHAR
typedef std::wstring String;
typedef wchar_t Char;
#else
typedef std::string String;
typedef char Char;
#endif

enum TextTokenState {
	TEXT = 0,
	TAG = 1,
};

class Font
{
public:
	std::string name;
	int height;
	int strWidth(String &text) { return text.size(); };

	Font(std::string iname) : name(iname), height(10) {};
};

class TextStyle
{
public:
	int font;
	int color;
	bool active;
	long uri;
};


class TextSpan
{
public:
	String text;
	int x, y;
	int w, h;
	int line;  // FIXME: obsolete??
	int open;
	bool eol;
	//TextSpan* link_to_previous_for_multiline;
	// lists of opening and closing tags
	TextStyle openingStyle;
	TextStyle closingStyle;

	TextSpan() : text() {};
	TextSpan(TextSpan &span);
	void print();
};

TextSpan::TextSpan(TextSpan &span)
{
	TextSpan();
	span.openingStyle = openingStyle;
	span.closingStyle = openingStyle;
}

void TextSpan::print()
{
	printf ("Text: %ls\n", text.c_str());
	printf ("C: %06x; U: %ld; X: %d; Y: %d; O: %c; E: %c\n\n", openingStyle.color, openingStyle.active?openingStyle.uri:-1, x, y, open?'y':'n', eol?'y':'n');
}

class Text
{
	std::list<TextSpan*> spans;
	std::vector<Font*> fonts;
public:
	Text();
	~Text() {};
	void parseText(char* text);
	void removeLine ();
	void print();

	int width;
	int maxLines;

protected:
	TextTokenState state;

	bool escaped;
	String buffer;
	TextTokenState tokenType;
	String token;
	String* getToken();
	void parseTag (String &tag, String &name, String &param, bool &closing);
	void layout (std::list<TextSpan*>::iterator start);
	TextSpan* getParentSpan();
};

Text::Text() {
	width = 50;
	tokenType = TEXT;
	TextSpan *span = new TextSpan();
	span->open = 1;
	span->eol = false;
	spans.push_back(span);
	fonts.push_back(new Font("BASIC"));
}

String* Text::getToken()
{
	for (String::iterator it = buffer.begin(); it != buffer.end(); it++) {
		if (escaped) {
			token += *it;
			escaped = false;
		} else if (*it == '\\') {
			escaped = true;
		} else if (token.empty()) {
			tokenType = (*it == '[') ? TAG : TEXT;
			token += *it;
		} else if (tokenType == TEXT && *it == '[') {
			buffer.erase(buffer.begin(), it);
			return &token;
		} else if ((tokenType == TAG && *it == ']') || (tokenType == TEXT && *it == '\n')) {
			token += *it++;
			buffer.erase(buffer.begin(), it);
			return &token;
		} else {
			token += *it;
		}
	}

	buffer.clear();
	if (tokenType == TAG || token.empty())
		return NULL;
	else
		return &token;
}

void Text::parseTag (String &tag, String &name, String &param, bool &closing)
{
	int pos;
	if (tag[1] == '/')  {
		closing = true;
		pos = 2;
	} else {
		closing = false;
		pos = 1;
	}

	int eqpos = tag.find('=', pos);
	std::cout << "eqpos: " << eqpos << std::endl;
	if (eqpos >= 0) {
		name = tag.substr(pos, eqpos - pos);
		param = tag.substr(eqpos+1, tag.size()-eqpos-2);
	} else {
		name = tag.substr(pos, tag.size()-pos-1);
		param.clear();
	}

}

void Text::parseText(char* text)
{
#ifdef USE_WCHAR
	wchar_t mbtext[1000];
	mbstowcs (mbtext, text, 1000);
	buffer += mbtext;
#else
	buffer += text;
#endif

	std::list<TextSpan*>::iterator tail = spans.end();
	tail--;

	String *t = getToken ();
	while (t) {
		printf ("Token: %ls\n", t->c_str());
		//std::cout << "TTT " << *t << "\n";

		TextSpan *parent = getParentSpan ();
		TextSpan *last = spans.back();
		bool eol = false;

		if (tokenType == TEXT) {
			if (*(t->end()-1) == '\n') {
				t->erase (t->end()-1, t->end());
				eol = true;
			}

			if (last != parent || last == spans.front() || last->eol) {
				last = new TextSpan();
				last->open = 0;
				spans.push_back(last);
			}

			// text spans should be closed, so that won't work ...
			last->text += *t;
			last->eol = eol;
		} else { // TAG
			String name, param;
			bool closing=false;
			parseTag (*t, name, param, closing);
			printf ("Tag: %ls : %ls C=%c\n", name.c_str(), param.c_str(), closing?'y':'n');

			if (closing) {
				TextSpan *parent = getParentSpan();
				if (parent != spans.front()) {
					parent->open--;
				}
			} else {
				TextSpan *parent = getParentSpan ();
				TextSpan *last = spans.back();

				if (parent != last || !last->text.empty() || last == spans.front()) {
					last = new TextSpan(*parent);
					spans.push_back(last);
				}
				last->open++;
				// add to it
				// FIXME: we could use switch instead ...
				if (name[0] == 'C' || name[0] == 'c') {
					last->openingStyle.color = wcstoul(param.c_str(), NULL, 16);
				} else if (name[0] == 'S' || name[0] == 's') {
					last->openingStyle.active = true;
					last->openingStyle.uri = wcstoul(param.c_str(), NULL, 10);
				} else if (name[0] == 'F' || name[0] == 'f') {
					last->openingStyle.font = wcstoul(param.c_str(), NULL, 10);
				}
			}
		}

		t->clear();
		t = getToken ();
	}

	tail++;
	layout (tail);
}

void Text::layout (std::list<TextSpan*>::iterator start)
{
	std::list<TextSpan*>::iterator prevIter = start;
	prevIter--;

	while (start != spans.end()) {
		std::cout << "l" << std::endl;

		TextSpan *prev = *prevIter;
		TextSpan *span = *start;

		assert (span->openingStyle.font < fonts.size());
		Font *font = fonts[span->openingStyle.font];

		int w = font->strWidth(span->text);

		int x = prev->x + prev->w;
		int y = prev->y;
		int line =prev->line;

		if (x+w > width || prev->eol) {
			y += font->height; // FIXME: actually we need the max height of fonts on this line, but ...
			x = 0;
			line++;
		}
		// FIXME: exclusion rects, span splitting, justification/alignment, tabs, bulletted/numbered lists

		span->x = x;
		span->y = y;
		span->w = w;
		span->h = font->height;
		span->line = line;

		start++;
		prevIter++;
	}
}

void Text::removeLine()
{
	std::list<TextSpan*>::iterator start = spans.begin();
	start++;
	int line = (*start)->line;

	// start: find first that has line >= 0 (fix the 0'th one!)
	while ((*start)->line == line) { // !bad
		if ((*start)->open == 0) {
			// remove it
		} else {
			// span->text = "";
			// span->line = -1;
		}

	}

}

TextSpan* Text::getParentSpan ()
{
	for (std::list<TextSpan*>::reverse_iterator it = spans.rbegin(); it != spans.rend(); it++) {
		if ((*it)->open) {
			return *it;
		}
	}
	return NULL;
}

void Text::print()
{
	int lastLine = 0;
	for (std::list<TextSpan*>::iterator it = spans.begin(), e=spans.end(); it != e; it++) {
		(*it)->print();
	}

	for (std::list<TextSpan*>::iterator it = spans.begin(), e=spans.end(); it != e; it++) {
		if ((*it)->line != lastLine) {
			printf("\n");
			lastLine =(*it)->line;
		}
		printf("<%ls>", (*it)->text.c_str());
	}
	printf("\n");
}

